/* This is from the BIND 4.9.4 release, modified to compile by itself */

/* Copyright (c) 1996 by Internet Software Consortium.
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND INTERNET SOFTWARE CONSORTIUM DISCLAIMS
 * ALL WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL INTERNET SOFTWARE
 * CONSORTIUM BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR
 * PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
 * ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 */

#if defined(LIBC_SCCS) && !defined(lint)
static char rcsid[] = "$Id$";
#endif				/* LIBC_SCCS and not lint */

#include <sys/param.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <string.h>
#include <errno.h>
#include <stdio.h>

#define	IN6ADDRSZ	16
#define	INT16SZ		 2

#ifndef	AF_INET6
#define	AF_INET6	AF_MAX+1	/* just to let this compile */
#endif

/*
 * WARNING: Don't even consider trying to compile this on a system where
 * sizeof(int) < 4.  sizeof(int) > 4 is fine; all the world's not a VAX.
 */

static const char *inet_ntop4(const u_char * src, char *dst, size_t size);
static const char *inet_ntop6(const u_char * src, char *dst, size_t size);

/* char *
 * inet_ntop(af, src, dst, size)
 *	convert a network format address to presentation format.
 * return:
 *	pointer to presentation format address (`dst'), or NULL (see errno).
 * author:
 *	Paul Vixie, 1996.
 */
const char *inet_ntop(af, src, dst, size)
int af;
const void *src;
char *dst;
size_t size;
{
    switch (af) {
    case AF_INET:
	return (inet_ntop4(src, dst, size));
    case AF_INET6:
	return (inet_ntop6(src, dst, size));
    default:
	errno = EAFNOSUPPORT;
	return (NULL);
    }
    /* NOTREACHED */
}

/* const char *
 * inet_ntop4(src, dst, size)
 *	format an IPv4 address, more or less like inet_ntoa()
 * return:
 *	`dst' (as a const)
 * notes:
 *	(1) uses no statics
 *	(2) takes a u_char* not an in_addr as input
 * author:
 *	Paul Vixie, 1996.
 */
static const char *inet_ntop4(src, dst, size)
const u_char *src;
char *dst;
size_t size;
{
    static const char fmt[] = "%u.%u.%u.%u";
    char tmp[sizeof "255.255.255.255"];

    sprintf(tmp, fmt, src[0], src[1], src[2], src[3]);
    if (strlen(tmp) > size) {
	errno = ENOSPC;
	return (NULL);
    }
    strcpy(dst, tmp);
    return (dst);
}

/* const char *
 * inet_ntop6(src, dst, size)
 *	convert IPv6 binary address into presentation (printable) format
 * author:
 *	Paul Vixie, 1996.
 */
static const char *inet_ntop6(src, dst, size)
const u_char *src;
char *dst;
size_t size;
{
    /* 
     * Note that int32_t and int16_t need only be "at least" large enough
     * to contain a value of the specified size.  On some systems, like
     * Crays, there is no such thing as an integer variable with 16 bits.
     * Keep this in mind if you think this function should have been coded
     * to use pointer overlays.  All the world's not a VAX.
     */
    char tmp[sizeof "ffff:ffff:ffff:ffff:ffff:ffff:255.255.255.255"], *tp;
    struct {
	int base, len;
    } best, cur;
    u_int words[IN6ADDRSZ / INT16SZ];
    int i;

    /* 
     * Preprocess:
     *      Copy the input (bytewise) array into a wordwise array.
     *      Find the longest run of 0x00's in src[] for :: shorthanding.
     */
    memset(words, 0, sizeof words);
    for (i = 0; i < IN6ADDRSZ; i++)
	words[i / 2] |= (src[i] << ((1 - (i % 2)) << 3));
    best.base = -1;
    cur.base = -1;
    for (i = 0; i < (IN6ADDRSZ / INT16SZ); i++) {
	if (words[i] == 0) {
	    if (cur.base == -1)
		cur.base = i, cur.len = 1;
	    else
		cur.len++;
	} else {
	    if (cur.base != -1) {
		if (best.base == -1 || cur.len > best.len)
		    best = cur;
		cur.base = -1;
	    }
	}
    }
    if (cur.base != -1) {
	if (best.base == -1 || cur.len > best.len)
	    best = cur;
    }
    if (best.base != -1 && best.len < 2)
	best.base = -1;

    /* 
     * Format the result.
     */
    tp = tmp;
    for (i = 0; i < (IN6ADDRSZ / INT16SZ); i++) {
	/* Are we inside the best run of 0x00's? */
	if (best.base != -1 && i >= best.base && i < (best.base + best.len)) {
	    if (i == best.base)
		*tp++ = ':';
	    continue;
	}
	/* Are we following an initial run of 0x00s or any real hex? */
	if (i != 0)
	    *tp++ = ':';
	/* Is this address an encapsulated IPv4? */
	if (i == 6 && best.base == 0 && (best.len == 6 || (best.len == 5 && words[5] == 0xffff))) {
	    if (!inet_ntop4(src + 12, tp, sizeof tmp - (tp - tmp)))
		return (NULL);
	    tp += strlen(tp);
	    break;
	}
	sprintf(tp, "%x", words[i]);
	tp += strlen(tp);
    }
    /* Was it a trailing run of 0x00's? */
    if (best.base != -1 && (best.base + best.len) == (IN6ADDRSZ / INT16SZ))
	*tp++ = ':';
    *tp++ = '\0';

    /* 
     * Check for overflow, copy, and we're done.
     */
    if ((tp - tmp) > size) {
	errno = ENOSPC;
	return (NULL);
    }
    strcpy(dst, tmp);
    return (dst);
}
